SET NOCOUNT ON;
go
USE ImEx;
go
IF EXISTS
 (SELECT *
  FROM   sys.objects
  WHERE  object_id = OBJECT_ID(N'dbo.FillThenNext')
  AND    type IN (N'P', N'PC'))
BEGIN;
  DROP PROCEDURE dbo.FillThenNext;
END;
go
CREATE PROCEDURE dbo.FillThenNext
                 @MaxCandidatesPerSession smallint
AS
-- Declare variables to hold fields read from cursor
DECLARE @Year          smallint
       ,@Quarter       tinyint
       ,@SubjCode      char(3)
       ,@NumCandidates smallint
-- Declare variables to detect year/quarter change
       ,@CurYear       smallint
       ,@CurQuarter    tinyint
-- Declare variables for session currently being filled and space left in it
       ,@CurSessionNo  smallint
       ,@CurSpaceLeft  smallint;

-- Init year/quarter change variables
SET @CurYear = 0;

-- Declare and open cursor to read registrations
DECLARE RegistrationsCursor CURSOR LOCAL FORWARD_ONLY STATIC READ_ONLY
FOR SELECT   Year, Quarter, SubjCode, NumCandidates
    FROM     dbo.Registrations
    ORDER BY Year, Quarter;

OPEN RegistrationsCursor;

-- Fetch first row
FETCH NEXT
FROM  RegistrationsCursor
INTO  @Year, @Quarter, @SubjCode, @NumCandidates;

-- Process all rows
WHILE @@FETCH_STATUS = 0
BEGIN;

  -- Check for change of year or quarter (includes first row of first quarter)
  IF @Year    <> @CurYear
  OR @Quarter <> @CurQuarter
  BEGIN;
    -- Write current session (if any) to disk
    IF @CurSessionNo IS NOT NULL
    BEGIN;
      INSERT INTO dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
      VALUES (@CurYear, @CurQuarter, @CurSessionNo, @CurSpaceLeft);
    END;
    -- Open first session in new quarter as current session
    SET @CurYear      = @Year;
    SET @CurQuarter   = @Quarter;
    SET @CurSessionNo = 1;
    SET @CurSpaceLeft = @MaxCandidatesPerSession;
  END;

  -- Check if current session has enough room for registration
  IF @CurSpaceLeft < @NumCandidates
  BEGIN;
    -- Not enough room - write current session to disk ...
    INSERT INTO dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
    VALUES (@CurYear, @CurQuarter, @CurSessionNo, @CurSpaceLeft);

    -- ... and open a new session
    SET @CurSessionNo = @CurSessionNo + 1;
    SET @CurSpaceLeft = @MaxCandidatesPerSession;
  END;

  -- Add registration to current session and decrement space left
  UPDATE dbo.Registrations
  SET    SessionNo = @CurSessionNo
  WHERE  Year      = @Year
  AND    Quarter   = @Quarter
  AND    SubjCode  = @SubjCode;

  SET @CurSpaceLeft = @CurSpaceLeft - @NumCandidates;

  -- Fetch next row
  FETCH NEXT
  FROM  RegistrationsCursor
  INTO  @Year, @Quarter, @SubjCode, @NumCandidates;
END;

-- Write last current session (if any) to disk
IF @CurSessionNo IS NOT NULL
BEGIN;
  INSERT INTO dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
  VALUES (@CurYear, @CurQuarter, @CurSessionNo, @CurSpaceLeft);
END;

-- Clean up resources used by cursor
CLOSE      RegistrationsCursor;
DEALLOCATE RegistrationsCursor;
go
