SET NOCOUNT ON;
go
USE ImEx;
go
IF EXISTS
 (SELECT *
  FROM   sys.objects
  WHERE  object_id = OBJECT_ID(N'dbo.FillThenSearch')
  AND    type IN (N'P', N'PC'))
BEGIN;
  DROP PROCEDURE dbo.FillThenSearch;
END;
go
CREATE PROCEDURE dbo.FillThenSearch
                 @MaxCandidatesPerSession smallint
AS
-- Declare variables to hold fields read from cursor
DECLARE @Year          smallint
       ,@Quarter       tinyint
       ,@SubjCode      char(3)
       ,@NumCandidates smallint
-- Declare variables to detect year/quarter change
       ,@CurYear       smallint
       ,@CurQuarter    tinyint
-- Declare variables for session currently being filled and space left in it
       ,@CurSessionNo  smallint
       ,@CurSpaceLeft  smallint
-- And finally a variable for the session to add a registration to
       ,@SessionNo     smallint;

-- Init year/quarter change variables
SET @CurYear = 0;

-- Declare and open cursor to read registrations
DECLARE RegistrationsCursor CURSOR LOCAL FORWARD_ONLY STATIC READ_ONLY
FOR SELECT   Year, Quarter, SubjCode, NumCandidates
    FROM     dbo.Registrations
    ORDER BY Year, Quarter;

OPEN RegistrationsCursor;

-- Fetch first row
FETCH NEXT
FROM  RegistrationsCursor
INTO  @Year, @Quarter, @SubjCode, @NumCandidates;

-- Process all rows
WHILE @@FETCH_STATUS = 0
BEGIN;

  -- Check for change of year or quarter (includes first row of first quarter)
  IF @Year    <> @CurYear
  OR @Quarter <> @CurQuarter
  BEGIN;
    -- Write last "current" session of previous quarter (if any) to disk
    IF @CurSessionNo IS NOT NULL
    BEGIN;
      INSERT INTO dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
      VALUES (@CurYear, @CurQuarter, @CurSessionNo, @CurSpaceLeft);
    END;
    -- Open first session in new quarter as "current" session
    SET @CurYear      = @Year;
    SET @CurQuarter   = @Quarter;
    SET @CurSessionNo = 1;
    SET @CurSpaceLeft = @MaxCandidatesPerSession;
  END;

  -- Check if "current" session has enough room for registration
  IF @CurSpaceLeft >= @NumCandidates
  BEGIN;
    -- It has, so registration can be added here
    SET @SessionNo = @CurSessionNo;
  END;
  ELSE
  BEGIN;
    -- Not enough room - check for another existing session with enough room
    SET @SessionNo = (SELECT TOP 1 SessionNo
                      FROM   dbo.Sessions
                      WHERE  Year       = @Year
                      AND    Quarter    = @Quarter
                      AND    SpaceLeft >= @NumCandidates);

    -- If no other session found, we'll have to make a new one
    IF @SessionNo IS NULL
    BEGIN;
      -- First write the former "current" session to disk ...
      INSERT INTO dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
      VALUES (@CurYear, @CurQuarter, @CurSessionNo, @CurSpaceLeft);

      -- ... then create a new "current" session
      SET @CurSessionNo = @CurSessionNo + 1;
      SET @CurSpaceLeft = @MaxCandidatesPerSession;
      -- Registration can be added to new "current" session
      SET @SessionNo = @CurSessionNo;
    END;
  END;

  -- Add registration to selected session
  UPDATE dbo.Registrations
  SET    SessionNo = @SessionNo
  WHERE  Year      = @Year
  AND    Quarter   = @Quarter
  AND    SubjCode  = @SubjCode;

  -- Decrement space left in selected session
  IF @CurSessionNo = @SessionNo
  BEGIN;
    -- "Current" session not yet written to disk - update variable
    SET @CurSpaceLeft = @CurSpaceLeft - @NumCandidates;
  END;
  ELSE
  BEGIN;
    -- Other sessions are on disk, so use UPDATE statement
    UPDATE dbo.Sessions
    SET    SpaceLeft = SpaceLeft - @NumCandidates
    WHERE  Year      = @Year
    AND    Quarter   = @Quarter
    AND    SessionNo = @SessionNo;
  END;

  -- Fetch next row
  FETCH NEXT
  FROM  RegistrationsCursor
  INTO  @Year, @Quarter, @SubjCode, @NumCandidates;
END;

-- Write last "current" session (if any) to disk
IF @CurSessionNo IS NOT NULL
BEGIN;
  INSERT INTO dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
  VALUES (@CurYear, @CurQuarter, @CurSessionNo, @CurSpaceLeft);
END;

-- Clean up resources used by cursor
CLOSE      RegistrationsCursor;
DEALLOCATE RegistrationsCursor;
go
